package cn.com.qmth.examcloud.core.oe.student.face.starter.config;

import cn.com.qmth.examcloud.commons.exception.StatusException;
import cn.com.qmth.examcloud.commons.helpers.concurrency.simple.ConcurrentTask;
import cn.com.qmth.examcloud.commons.util.Util;
import cn.com.qmth.examcloud.core.oe.common.entity.ExamCaptureQueueEntity;
import cn.com.qmth.examcloud.core.oe.common.repository.ExamCaptureQueueRepo;
import cn.com.qmth.examcloud.core.oe.student.face.service.impl.BaiduFaceLivenessWorker;
import cn.com.qmth.examcloud.core.oe.student.face.service.impl.ExamCaptureProcessStatisticController;
import cn.com.qmth.examcloud.exchange.inner.api.SmsCloudService;
import cn.com.qmth.examcloud.exchange.inner.api.request.SendSmsReq;
import cn.com.qmth.examcloud.support.cache.CacheHelper;
import cn.com.qmth.examcloud.support.cache.bean.SysPropertyCacheBean;
import cn.com.qmth.examcloud.web.bootstrap.PropertyHolder;
import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.boot.ApplicationArguments;
import org.springframework.boot.ApplicationRunner;
import org.springframework.core.annotation.Order;
import org.springframework.stereotype.Component;

import java.util.Arrays;
import java.util.List;
import java.util.Timer;
import java.util.TimerTask;

/**
 * 启动百度活体检测任务
 */
@Component
@Order(50)
public class ProcessBaiduFacelivenessTask implements ApplicationRunner {

    private static final Logger log = LoggerFactory.getLogger(ProcessBaiduFacelivenessTask.class);
    @Autowired
    ExamCaptureQueueRepo examCaptureQueueRepo;
    private final Log captureLog = LogFactory.getLog("PROCESS_EXAM_CAPTURE_TASK_LOGGER");
    private final ExamCaptureProcessStatisticController statisticController = new ExamCaptureProcessStatisticController();
    //失败率预警阈值
    private static final double RATE_WARN_THRESHOLD = 0.5;
    //总数量预警阈值
    private static final int COUNT_WARN_THRESHOLD = 10;
    @Autowired
    SmsCloudService smsCloudService;

    private void start() {
        sendAlarmSmsIfNecessary();

        ConcurrentTask concurrentTask = new ConcurrentTask();
        concurrentTask.setMaxActiveThreadSize(PropertyHolder.getInt("$capture.thread.maxActiveThreadSize", 100));
        concurrentTask.setMinThreadSize(PropertyHolder.getInt("$capture.thread.minThreadSize", 2));
        concurrentTask.setWorker(new BaiduFaceLivenessWorker(statisticController));
        concurrentTask.start();
        //当前获取数据的批次号(默认用时间戳)
        String processBatchNum = "B_" + System.currentTimeMillis();

        captureLog.debug("[PROCESS_BAIDUFACELIVENESS." + processBatchNum + "] 启动百度处理服务...");
        while (true) {
            try {
                //如果队列没满，则从数据库中查数据并插入
                List<ExamCaptureQueueEntity> examCaptureQueueList = examCaptureQueueRepo.
                        findNeedFacelivenessDetectExamCaptureQueuesLimit(PropertyHolder.getInt("$capture.queue.limit", 100), processBatchNum);
                //如果队列中没取到数据，则2秒钟后再取
                if (null == examCaptureQueueList || examCaptureQueueList.isEmpty()) {
                    captureLog.debug("[PROCESS_BAIDUFACELIVENESS." + processBatchNum + "] 抓拍队列中没有取到数据，2秒后重试");

                    Util.sleep(PropertyHolder.getInt("$capture.queue.read.sleepSeconds.", 2));
                    continue;
                }

                captureLog.debug("[PROCESS_BAIDUFACELIVENESS." + processBatchNum + "] 抓拍队列中的数据条数为：" + examCaptureQueueList.size());

                for (ExamCaptureQueueEntity offeredQueueEntity : examCaptureQueueList) {
                    while (true) {
                        boolean offerSuccess = concurrentTask.offerElement(offeredQueueEntity);
                        if (offerSuccess) {
                            offeredQueueEntity.setProcessBatchNum(processBatchNum);
                            examCaptureQueueRepo.save(offeredQueueEntity);

                            captureLog.debug("[PROCESS_BAIDUFACELIVENESS." + processBatchNum + "] 向工作队列中添加数据成功：fileUrl=" + offeredQueueEntity.getFileUrl());
                            break;
                        }

                        captureLog.debug("[PROCESS_BAIDUFACELIVENESS." + processBatchNum + "] 向工作队列中添加数据失败，30秒后重试：fileUrl=" + offeredQueueEntity.getFileUrl());

                        Util.sleep(PropertyHolder.getInt("$capture.queue.offer.sleepSeconds.", 30));
                    }
                }
            } catch (Exception e) {
                log.error("300002", "启动百度活体检测队列出现异常", e);
                captureLog.error("[PROCESS_BAIDUFACELIVENESS." + processBatchNum + "] 百度活体检测出出异常,3秒后重试", e);
                Util.sleep(3);
            }
        }
    }

    /**
     * 如果有必要则发短信
     */
    private void sendAlarmSmsIfNecessary() {
        //每分钟判断是否需要发短信报警
        Timer timer = new Timer();
        timer.scheduleAtFixedRate(new TimerTask() {
            @Override
            public void run() {
                //如果每分钟失败率超过50%则发短信报警，且总数不少于10次则短信报警
                if (statisticController.getCount() > COUNT_WARN_THRESHOLD &&
                        statisticController.getFailureRate() > RATE_WARN_THRESHOLD) {
                    SysPropertyCacheBean smsPhoneProperty = CacheHelper.getSysProperty("capture.sms.phones");
                    if (!smsPhoneProperty.getHasValue()) {
                        throw new StatusException("300001", "未配置图片处理失败的通知手机号");
                    }
                    List<String> phoneList = Arrays.asList(smsPhoneProperty.getValue().toString().split(","));
                    SendSmsReq sendSmsReq = new SendSmsReq();
                    sendSmsReq.setPhoneList(phoneList);
                    sendSmsReq.setSmsAssemblyCode("CAPTURE");
                    sendSmsReq.setParams(null);
                    smsCloudService.sendSms(sendSmsReq);
                }
                //每1分钟重置一次总数量与失败数量
                statisticController.resetCount();
                statisticController.resetFailureCount();
            }
        }, 100, 60 * 1000);
    }

    @Override
    public void run(ApplicationArguments args) {
        new Thread(() -> {
            start();
        }).start();
    }

}
